# Generated by Django 4.2.6 on 2023-12-01 17:17

from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ("nvc_dhl_tracker", "0002_dhlpackage_dhlapikey_exhausted"),
    ]

    operations = [
        migrations.AddField(
            model_name="dhlpackage",
            name="origin",
            field=models.CharField(default="", max_length=1024),
            preserve_default=False,
        ),
    ]



import itertools

from .dhl import query_and_parse_responses
from .google import update_google_sheet


async def track(google_sheet_url, tracking_sheet_id, meta_sheet_id):
    jsons = await query_and_parse_responses(
        dhl_numbers=(f"JD0030110676223{n:03d}" for n in range(10)),
        delay=1.0,
    )
    update_google_sheet(
        google_sheet_url,
        tracking_sheet_id,
        meta_sheet_id,
        itertools.chain(*jsons),
    )



# Generated by Django 4.2.6 on 2023-12-01 18:40

from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ("nvc_dhl_tracker", "0005_alter_dhlpackage_number"),
    ]

    operations = [
        migrations.AlterField(
            model_name="dhlpackage",
            name="became_delivered",
            field=models.DateTimeField(null=True),
        ),
        migrations.AlterField(
            model_name="dhlpackage",
            name="became_pre_transit",
            field=models.DateTimeField(null=True),
        ),
        migrations.AlterField(
            model_name="dhlpackage",
            name="became_transit",
            field=models.DateTimeField(null=True),
        ),
    ]



# Generated by Django 4.2.6 on 2023-12-01 17:20

from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ("nvc_dhl_tracker", "0004_alter_dhlpackage_discovered"),
    ]

    operations = [
        migrations.AlterField(
            model_name="dhlpackage",
            name="number",
            field=models.CharField(max_length=1024, unique=True),
        ),
    ]



from typing import Iterable

import gspread

from i129f.nvc_dhl_tracker.models import DhlPackage, GoogleKeyUsage
from i129f.nvc_dhl_tracker.signals.sender import google_request_made


def update_google_sheet(sheet_key, sheet, dhl_packages: Iterable[DhlPackage]):
    gc = gspread.service_account()
    sh = gc.open_by_url(sheet_key)
    table = {
        "headers": [
            "Tracking Number",
            "Discovered",
            "Location",
            "Status",
            "Status - Pre Transit",
            "Status - Transit",
            "Status - Delivered",
            "Total Time in Transit (seconds)",
        ],
    }

    for package in dhl_packages:
        table[package.number] = [
            package.number,
            package.created.isoformat(),
            package.location,
            package.current_status,
            package.became_pre_transit.isoformat(),
            package.became_transit.isoformat(),
            package.became_delivered.isoformat(),
            (package.became_delivered - package.became_pre_transit).total_seconds(),
        ]

    sh.get_worksheet_by_id(sheet).update(range_name="A1", values=list(table.values()))
    google_request_made.send(sender=GoogleKeyUsage)



# Generated by Django 4.2.6 on 2023-12-01 17:06

from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ("nvc_dhl_tracker", "0001_initial"),
    ]

    operations = [
        migrations.CreateModel(
            name="DhlPackage",
            fields=[
                (
                    "id",
                    models.BigAutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("created", models.DateTimeField(auto_now_add=True)),
                ("updated", models.DateTimeField(auto_now=True)),
                ("number", models.CharField(max_length=1024)),
                ("discovered", models.DateTimeField()),
                ("location", models.CharField(max_length=1024)),
                ("current_status", models.CharField(max_length=512)),
                ("became_pre_transit", models.DateTimeField()),
                ("became_transit", models.DateTimeField()),
                ("became_delivered", models.DateTimeField()),
            ],
        ),
        migrations.AddField(
            model_name="dhlapikey",
            name="exhausted",
            field=models.BooleanField(default=False),
        ),
    ]



import asyncio
import logging
import urllib.parse
from json import JSONDecodeError
from typing import Iterable

import httpx

from i129f.nvc_dhl_tracker.models import DhlApiKey
from i129f.nvc_dhl_tracker.signals.sender import (
    post_dhl_request_made,
    pre_dhl_request_made,
)

log = logging.getLogger(__name__)

host = "api-eu.dhl.com"
path = "/track/shipments"


async def query_dhl_number(
    client: httpx.AsyncClient, dhl_number: str, instance: DhlApiKey
) -> httpx.Response:
    query = urllib.parse.urlencode(
        {
            "trackingNumber": dhl_number,
        }
    )
    headers = {
        "Accept": "application/json",
        "DHL-API-Key": instance.key,
    }
    url = urllib.parse.urlunsplit(("https", host, path, "", ""))

    log.info("Querying DHL for %s at %s", dhl_number, url)
    try:
        await instance.can_make_request()
        pre_dhl_request_made.send_robust(DhlApiKey, instance=instance)
        r = await client.get(url, headers=headers, params=query)
        try:
            json = r.json()
        except JSONDecodeError:
            json = {}

        post_dhl_request_made.send_robust(
            DhlApiKey, instance=instance, response=r, json=json
        )

        log.info("DHL returned a HTTP-%s for %s", r.status_code, dhl_number)
    except Exception as e:
        log.exception(
            "Failed to query DHL at %s for %s because of %s(%s)",
            url,
            dhl_number,
            type(e),
            str(e),
        )
    else:
        return r


async def query_dhl_numbers(
    dhl_numbers: Iterable[str],
    client: httpx.AsyncClient,
    instance: DhlApiKey,
    delay: float = 0.0,
):
    for dhl_number in dhl_numbers:
        tracking_information = await query_dhl_number(client, dhl_number, instance)
        if tracking_information:
            yield dhl_number, tracking_information
        else:
            log.info("Skipping %s", dhl_number)

        if delay > 0:
            log.info("Sleeping for %s", delay)
            await asyncio.sleep(delay)


def divine_statuses(events):
    for event in reversed(events):
        pre_transit_found = False
        transit_found = False
        if "Shipment picked up" in event["description"] and not pre_transit_found:
            pre_transit_found = True
            yield "became_pre_transit", event["timestamp"]
        elif "Shipment has departed" in event["description"] and not transit_found:
            transit_found = True
            yield "became_transit", event["timestamp"]
        elif "Delivered" in event["description"]:
            yield "became_delivered", event["timestamp"]
            break


def retrieve_status(dhl_number, as_json):
    try:
        yield from parse_json(as_json)
    except KeyError:
        log.info("%s was not found in the DHL API", dhl_number)
        yield dhl_number, {
            "status": "not_found",
            "location": "UNKNOWN",
        }


def parse_json(as_json):
    for shipment in as_json["shipments"]:
        for piece in shipment["details"]["pieceIds"]:
            status = shipment["status"]
            yield piece, {
                "status": status["statusCode"],
                "location": status["location"]["address"]["addressLocality"],
                "origin": shipment["origin"]["address"]["addressLocality"],
                "destination": shipment["destination"]["address"]["addressLocality"],
                "history": [
                    {
                        "timestamp": event["timestamp"],
                        "location": event["location"]["address"]["addressLocality"],
                    }
                    for event in shipment["events"]
                ],
                "statuses": dict(divine_statuses(shipment["events"])),
            }


def convert_to_json(dhl_number, response):
    try:
        as_json = response.json()
        return retrieve_status(dhl_number, as_json)
    except Exception:
        log.exception(
            "Unexpected exception while converting to Json. %s -> %s.",
            response.request.url,
            response.status_code,
        )


async def query_and_parse_responses(
    dhl_numbers: Iterable[str], delay: float, instance: DhlApiKey
):
    async with httpx.AsyncClient() as client:
        responses = [
            response
            async for response in query_dhl_numbers(
                dhl_numbers, client, instance, delay
            )
        ]
        return (convert_to_json(*response) for response in responses)



import asyncio
import logging

from django.dispatch import receiver

from i129f.nvc_dhl_tracker.dhl import parse_json
from i129f.nvc_dhl_tracker.models import (
    DhlApiKey,
    DhlKeyUsage,
    DhlPackage,
    GoogleKeyUsage,
)
from i129f.nvc_dhl_tracker.signals.sender import (
    google_request_made,
    post_dhl_request_made,
    pre_dhl_request_made,
)


@receiver(google_request_made, sender=GoogleKeyUsage)
def record_google_api_request(**kwargs):
    asyncio.create_task(GoogleKeyUsage.objects.acreate())


@receiver(pre_dhl_request_made, sender=DhlApiKey)
def record_dhl_api_request(instance, **kwargs):
    asyncio.create_task(DhlKeyUsage.objects.acreate(key=instance))


@receiver(post_dhl_request_made, sender=DhlApiKey)
def record_dhl_package(instance, response, json, **kwargs):
    if json and response.is_success:
        try:
            data = list(parse_json(json))
        except Exception as e:
            logging.exception("Could not parse %s because %s(%s)", json, type(e), e)
            return

        for number, package_data in data:
            asyncio.create_task(
                DhlPackage.objects.aupdate_or_create(
                    number=number,
                    defaults=dict(
                        origin=package_data["origin"],
                        location=package_data["location"],
                        current_status=package_data["status"],
                        **package_data["statuses"]
                    ),
                ),
                name="Save DHL Package Data",
            )
    else:
        if response.status_code in {401, 403}:
            instance.exhausted = True
            instance.save()
            logging.error("Permission denied, exhausting API key!")



# Generated by Django 4.1 on 2023-12-01 16:32

import django.db.models.deletion
from django.db import migrations, models


class Migration(migrations.Migration):
    initial = True

    dependencies = []

    operations = [
        migrations.CreateModel(
            name="DhlApiKey",
            fields=[
                (
                    "id",
                    models.BigAutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("created", models.DateTimeField(auto_now_add=True)),
                ("updated", models.DateTimeField(auto_now=True)),
                ("key", models.CharField(max_length=1024)),
            ],
        ),
        migrations.CreateModel(
            name="GoogleKeyUsage",
            fields=[
                (
                    "id",
                    models.BigAutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("created", models.DateTimeField(auto_now_add=True)),
                ("updated", models.DateTimeField(auto_now=True)),
            ],
        ),
        migrations.CreateModel(
            name="DhlKeyUsage",
            fields=[
                (
                    "id",
                    models.BigAutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("created", models.DateTimeField(auto_now_add=True)),
                ("updated", models.DateTimeField(auto_now=True)),
                (
                    "key",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        to="nvc_dhl_tracker.dhlapikey",
                    ),
                ),
            ],
        ),
    ]



# Generated by Django 4.2.6 on 2023-12-01 17:19

from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ("nvc_dhl_tracker", "0003_dhlpackage_origin"),
    ]

    operations = [
        migrations.AlterField(
            model_name="dhlpackage",
            name="discovered",
            field=models.DateTimeField(auto_now_add=True),
        ),
    ]
